//Ethan Alexander Shulman 2023
/*File: GL.js
WebGL 1 wrapper, defines object classes for OpenGL objects. RenderTexture = framebuffer, Shader = program, Mesh = vertex/index buffer.*/

/*Property: gl
*WebGLRenderingContext* The WebGL rendering context for displayCanvas.*/
var gl,
/*Property: renderingCanvas
*HTMLCanvasElement* The canvas used for WebGL rendering.*/
renderingCanvas,
/*Property: pixelScaling
*float* DPI scaling value.*/
pixelScaling = 1,
/*Property: display
<RenderTexture> used to render to the display.*/
display = {framebuffer:null,width:0,height:0},
/*Property: boundTextures
*Array* Of uniform locations corresponding to currently bound textures, used to automatically unbind textures.*/
boundTextures = [],
/*Property: shaders
*Array* Of all <Shader>'s.*/
shaders = [],
/*Property: currentShader
<Shader> Currently bound shader.*/
currentShader = null,
/*Property: currentMesh
<Mesh> Currently bound mesh.*/
currentMesh = null,
/*Property: currentRenderTexture
<RenderTexture> Currently bound render texture, defaults to <display>.*/
currentRenderTexture = null,
/*Property: currentRenderTextureX/Y/W/H
*float* X, y, width and height coordinates describing currently bound viewport dimensions.*/
currentRenderTextureX, currentRenderTextureY, currentRenderTextureW, currentRenderTextureH,
/*Property: fullscreenTriangle
<Mesh> Fullscreen triangle mesh.*/
fullscreenTriangle,
/*Property: fullscreenQuad
<Mesh> Fullscreen quad mesh.*/
fullscreenQuad,

//curent number of enabled vertex attributes
currentVertexAttribs = 0,
//true if currentMesh vertex/index buffer needs to be rebound
currentMeshVBChanged = false,
currentMeshIBChanged = false,

/*Property: SHADER_HEADER
*String* Shader header text with #version and precision highp declarations.*/
SHADER_HEADER = "#version 100\nprecision highp float;precision highp int;\n";



/*Function: InitializeGraphics
Initialize graphics system.

Parameter:
*HTMLCanvasElement* canvas - Display canvas to render to.
*bool* msaa - Anti aliasing.

Returns:
*bool* True if success, false if failed.*/
function InitializeGraphics() {
	renderingCanvas = document.getElementById("GLCanvas");
	
	//create webgl2 context
	var gopt = {"alpha":true, "antialias":false, "depth":false, "failIfMajorPerformanceCaveat":false, "preserveDrawingBuffer":false};
	gl = renderingCanvas.getContext("webgl",gopt);
	if (!gl) gl = renderingCanvas.getContext("experimental-webgl",gopt);
	if (!gl) {
		FatalCrash("Fatal Error: WebGL not supported by your browser or device.");
		return false;
	}
		
	//default webgl state
	gl.enable(gl.CULL_FACE);
	gl.cullFace(gl.BACK);
	gl.disable(gl.DEPTH_TEST);
	gl.depthMask(false);
	gl.disable(gl.STENCIL_TEST);
	gl.blendFunc(gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
	gl.disable(gl.BLEND);
	gl.enable(gl.SCISSOR_TEST);
	gl.colorMask(true,true,true,true);
	gl.pixelStorei(gl.UNPACK_ALIGNMENT,1);
	gl.pixelStorei(gl.PACK_ALIGNMENT,1);
	
	//create fullscreen triangle(triangles) and quad mesh(triangle fan)
	var fsAttribs = [new MeshAttrib("vPos",2,4,gl.FLOAT)];
	fullscreenTriangle = new Mesh(fsAttribs,new Float32Array([2,1, -2,1, 0,-3]));
	fullscreenQuad = new Mesh(fsAttribs,new Float32Array([1,1, -1,1, 1,-1, -1,-1]));
	
	return true;
}




/*Class: MeshAttrib
Mesh vertex attribute description.

Constructor: MeshAttrib
New MeshAttrib.

Parameters:
*String* name - Attribute name in shaders.
*int* count - Number of attribute components must be in range 1-4(inclusive).
*int* bytes - Bytes per component.
*WebGLConstant* type - gl.FLOAT,gl.INT,etc.*/
/**@constructor*/
function MeshAttrib(name,count,bytes,type) {
	/*Property: count*/
	this.count = count;
	/*Property: bytes*/
	this.bytes = bytes;
	/*Property: size*/
	this.size = count*bytes;
	/*Property: type*/
	this.type = type;
	/*Property: name*/
	this.name = name;
}

/*Class: Mesh
Mesh handles vertex data on the GPU for rendering.

Constructor: Mesh
New Mesh.

Parameters:
*Array* of <MeshAttrib> attribs - Vertex attributes.
(optional) *TypedArray* data - Interleaved vertex data.
(optional) *WebGLConstant* usage - Defaults to gl.STATIC_DRAW.*/
/**@constructor*/
function Mesh(attribs, data, usage) {
	/*Var: attributes*/
	this.attributes = attribs;
	/*Var: buffer*/
	this.buffer = gl.createBuffer();
	/*Var: indexBuffer
	*WebGLBuffer* Index element buffer.*/
	this.indexBuffer = null;
	/*Var: vertexCount*/
	this.vertexCount = 0;
	/*Var: indexCount*/
	this.indexCount = 0;
	
	this.vertexData = null;
	this.indexData = null;
	
	var offset = 0;
	for (var i = 0; i < attribs.length; i++) offset += attribs[i].size;
	this.vertexSize = offset;
	
	if (data) {
		gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
		gl.bufferData(gl.ARRAY_BUFFER,data,usage?usage:gl.STATIC_DRAW);
		this.vertexCount = (data.length*data.BYTES_PER_ELEMENT)/this.vertexSize;
		if (currentMesh) gl.bindBuffer(gl.ARRAY_BUFFER, currentMesh.buffer);
	}
}

/*Function: Set
Set vertex data.

Parameters:
*TypedArray* data - Interleaved vertex data.
(optional) *WebGLConstant* usage - Defaults to gl.STATIC_DRAW.*/
Mesh.prototype.Set = function(data, usage) {
	if (currentMesh !== this) {
		gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
		currentMeshVBChanged = true;
	}
	this.vertexCount = (data.length*data.BYTES_PER_ELEMENT)/this.vertexSize;
	gl.bufferData(gl.ARRAY_BUFFER,data,usage?usage:gl.STATIC_DRAW);
}

/*Function: SetIndices
Set mesh index data.

Parameters:
*Uint16Array* indices - Index data.
*GLenum* usage - Either gl.STATIC_DRAW, gl.STREAM_DRAW or gl.DYNAMIC_DRAW, optional if undefined defaults to gl.STATIC_DRAW.*/
Mesh.prototype.SetIndices = function(indices,usage) {
	if (this.indexBuffer) {
		if (currentMesh !== this) {
			gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER,this.indexBuffer);
			currentMeshIBChanged = true;
		}
	} else {
		this.indexBuffer = gl.createBuffer();
		gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER,this.indexBuffer);
		if (currentMesh !== this) {
			currentMeshIBChanged = true;
		}
	}
	
	this.indexCount = indices.length;
	gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, indices, usage?usage:gl.STATIC_DRAW);
}

/*Function: Bind
Bind as current mesh for rendering.*/
Mesh.prototype.Bind = function() {
	if (currentMesh == this) {
		if (currentMeshVBChanged) gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
		if (currentMeshIBChanged) gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indexBuffer);
		return;
	}
	currentMesh = this;
	
	gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
	if (this.indexBuffer) gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indexBuffer);
	
	var attribs = this.attributes, offset = 0;
	for (var i = 0; i < attribs.length; i++) {
		var at = attribs[i];
		gl.vertexAttribPointer(i, at.count,at.type,false, at.size,offset);
		if (i >= currentVertexAttribs) gl.enableVertexAttribArray(i);
	}
	for (var i = attribs.length; i < currentVertexAttribs; i++) {
		gl.disableVertexAttribArray(i);
	}
	currentVertexAttribs = attribs.length;
}

/*Function: Delete
Delete mesh buffer.*/
Mesh.prototype.Delete = function() {
	gl.deleteBuffer(this.buffer);
	if (this.indexBuffer) gl.deleteBuffer(this.indexBuffer);
	if (currentMesh === this) currentMesh = null;
}




/*Function: CompileShader
Helper function for compiling vertex/fragment shaders.

Parameters:
*String* src - Shader source code.
*GLenum* type - Shader type, either gl.VERTEX_SHADER or gl.FRAGMENT_SHADER.

Returns:
*WebGLShader* or null if failed. Prints shader error and source to console on failure.*/
function CompileShader(src,type) {
	var s = gl.createShader(type);
	gl.shaderSource(s,src);
	gl.compileShader(s);
	if (!gl.getShaderParameter(s,gl.COMPILE_STATUS)) {
		console.error("Error compiling GLSL shader: "+gl.getShaderInfoLog(s)+"\nFrom shader:");
		console.log(src);
		gl.deleteShader(s);
		return null;
	}
	return s;
}


//load shader from text element and compile
function LoadShader(srcId, type) {
	return CompileShader(SHADER_HEADER+document.getElementById(srcId).textContent, type);
}


/*Class: Shader
Compiled shader program and uniform locations. You can access uniform locations by name for example if in shader there's 'uniform vec3 color' you can access the location via shader['color'].

Constructor: Shader
Construct new WebGL shader program.

Parameters:
*vshader* vert - Vertex shader.
*fshader* frag - Fragment shader.
(optional) *Array* of <MeshAttrib> meshAttribs - Mesh attributes to bind, if null binds 'vPos' to attribute 0.
*/
/**@constructor*/
function Shader(vert,frag,meshAttribs) {	
	var sp = gl.createProgram();
	gl.attachShader(sp,vert);
	gl.attachShader(sp,frag);
	if (meshAttribs) {
		for (var i = 0; i < meshAttribs.length; i++) gl.bindAttribLocation(sp,i,meshAttribs[i].name);
	} else {
		gl.bindAttribLocation(sp,0,"vPos");
	}
	gl.linkProgram(sp);
	gl.detachShader(sp,vert);
	gl.detachShader(sp,frag);
	if (!gl.getProgramParameter(sp,gl.LINK_STATUS)) {
		console.error("Error linking GLSL program: "+gl.getProgramInfoLog(sp));
		return;
	}
	
	var nuni = gl.getProgramParameter(sp, gl.ACTIVE_UNIFORMS);
	for (var i = 0; i < nuni; i++) {
		var u = gl.getActiveUniform(sp,i),
			as = u.name.indexOf("["),
			loc = gl.getUniformLocation(sp,u.name);
		if (loc) this[as===-1?u.name:u.name.substring(0,as)] = loc;
	}
	
	/*Property: program
	*WebGLProgram* Shader program.*/
	this.program = sp;
	
	shaders.push(this);
}

/*Function: Bind
Bind as current rendering shader.*/
Shader.prototype.Bind = function() {
	if (currentShader === this) return;
	currentShader = this;
	gl.useProgram(this.program);

}

/*Function: Delete
Delete shader program.*/
Shader.prototype.Delete = function() {
	gl.deleteProgram(this.program);
	shaders.splice(shaders.indexOf(this),1);
	if (currentShader === this) currentShader = null;
}


//load vert and frag shader from text elements and create new Shader
function LoadShaders(vertId, fragId, attribs) {
	var vs = LoadShader(vertId,gl.VERTEX_SHADER),
		fs = LoadShader(fragId,gl.FRAGMENT_SHADER),
		prog = new Shader(vs,fs, attribs);
	gl.deleteShader(vs);
	gl.deleteShader(fs);
	return prog;
}




/*Class: RenderTexture
Render texture manages framebuffer texture rendering.

Constructor: RenderTexture
New RenderTexture.

Parameters:
*int* w - Width in pixels.
*int* h - Height in pixels.*/
/**@constructor*/
function RenderTexture(w,h) {
	/*Property: framebuffer
	*WebGLFrameBuffer* Framebuffer object.*/
	this.framebuffer = gl.createFramebuffer();
	/*Property: texture
	*WebGLTexture* Texture attached to color attachment 0.*/
	this.texture = gl.createTexture();
	/*Property: width
	*int* Width in pixels.*/
	this.width = w;
	/*Property: height
	*int* Height in pixels.*/
	this.height = h;
	
	gl.bindTexture(gl.TEXTURE_2D,this.texture);
	gl.texImage2D(gl.TEXTURE_2D,0,gl.RGBA,w,h,0,gl.RGBA,gl.UNSIGNED_BYTE,null);
	gl.bindFramebuffer(gl.FRAMEBUFFER,this.framebuffer);
	gl.framebufferTexture2D(gl.FRAMEBUFFER,gl.COLOR_ATTACHMENT0,gl.TEXTURE_2D,this.texture,0);
	
	if (currentRenderTexture) gl.bindFramebuffer(gl.FRAMEBUFFER, currentRenderTexture.framebuffer);
}

/*Function: SetSize
Set size and optionally upload data.

Parameters:
*int* w - Width in pixels.
*int* h - Height in pixels.
(optional) dat - Image data to upload.*/
RenderTexture.prototype.SetSize = function(w,h,dat) {
	this.width = w;
	this.height = h;
	gl.bindTexture(gl.TEXTURE_2D,this.texture);
	gl.texImage2D(gl.TEXTURE_2D,0,gl.RGBA,w,h,0,gl.RGBA,gl.UNSIGNED_BYTE,dat);
}

/*Function: Bind
Bind as current render target.

Parameters:
(optional) *int* x - Destination left, min X. Defaults to 0.
(optional) *int* y - Destination bottom, min Y. Defaults to 0.
(optional) *int* w - Destination width, max X. Defaults to this.width.
(optional) *int* h - Destination, height, maxY. Defaults to this.height.*/
RenderTexture.prototype.Bind = function(x,y,w,h) {
	if (arguments.length < 4) {
		x = y = 0;
		w = this.width; h = this.height;
	}
	if (currentRenderTexture === this && currentRenderTextureX === x && currentRenderTextureY === y && currentRenderTextureW === w && currentRenderTextureH === h) return;
	currentRenderTexture = this;
	currentRenderTextureX = x; currentRenderTextureY = y;
	currentRenderTextureW = w; currentRenderTextureH = h;
	
	gl.bindFramebuffer(gl.FRAMEBUFFER,this.framebuffer);
	gl.viewport(x,y,w,h);
	gl.scissor(x,y,w,h);
}
display.Bind = RenderTexture.prototype.Bind;

/*Function: Delete
Delete render texture data.*/
RenderTexture.prototype.Delete = function() {
	gl.deleteFramebuffer(this.framebuffer);
	gl.deleteTexture(this.texture);
}




/*Function: SetTextureOptions
Set texture filtering options on currently bound texture 2d.

Parameters:
*bool* linear - If true sets linear filtering, if false nearest.
*bool* mipmap - If true enables mipmap filtering.
*bool* repeatx - If true enables x texture wrapping, if false clamping.
*bool* repeaty - If true enables y texture wrapping, if false clamping.*/
function SetTextureOptions(linear,mipmap,repeatx,repeaty) {
	var mgf, mif;
	if (linear) {
		mgf = gl.LINEAR;
		mif = mipmap?gl.LINEAR_MIPMAP_LINEAR:gl.LINEAR;
	} else {
		mgf = gl.NEAREST;
		mif = mipmap?gl.NEAREST_MIPMAP_LINEAR:gl.NEAREST;
	}
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_MAG_FILTER,mgf);
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_MIN_FILTER,mif);
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_WRAP_S,repeatx?gl.REPEAT:gl.CLAMP_TO_EDGE);
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_WRAP_T,repeaty?gl.REPEAT:gl.CLAMP_TO_EDGE);
}


/*Function: BindTexture
Bind texture.

Parameters:
*WebGLUniformLocation* uni - Uniform location to bind the texture to.
*WebGLTexture* tex - Texture to bind.
(optional) texType - Defauls to gl.TEXTURE_2D.*/
function BindTexture(uni,tex,texType) {
	var ni = boundTextures.indexOf(gl.INVALID_VALUE);
	if (ni === -1) ni = boundTextures.push(uni)-1;
	else boundTextures[ni] = uni;
	gl.activeTexture(gl.TEXTURE0+ni);
	gl.bindTexture(texType?texType:gl.TEXTURE_2D,tex);
	gl.uniform1i(uni,ni);
}

/*Function: UnbindTexture
Unbind texture.

Parameters:
*WebGLUniformLocation* uni - Uniform location texture was bound to.*/
function UnbindTexture(uni) {
	boundTextures[boundTextures.indexOf(uni)] = gl.INVALID_VALUE;
}

/*Function: UnbindAllTextures
Unbind all bound textures.*/
function UnbindAllTextures() {
	for (var i = 0; i < boundTextures.length; i++) boundTextures[i] = gl.INVALID_VALUE;
}


/*Function: SetDisplaySize
Set display canvas size, optionally correcting for DPI.

Parameters:
*int* w - Canvas width in pixels.
*int* h - Canvas height in pixels.
*bool* useDpi - If true will take DPI into account.*/
function SetDisplaySize(w,h,useDpi) {
	if (useDpi) {
		var ndpi = window.devicePixelRatio;
		if (!ndpi) ndpi = 1;
		renderingCanvas.width = display.width = Math.floor(w*ndpi);
		renderingCanvas.height = display.height = Math.floor(h*ndpi);
		if (ndpi !== 1 || pixelScaling !== 1) {
			renderingCanvas.style.width = w+"px";
			renderingCanvas.style.height = h+"px";
		}
		pixelScaling = ndpi;
		
	} else {
		renderingCanvas.width = display.width = w;
		renderingCanvas.height = display.height = h;
		if (pixelScaling !== 1) {
			renderingCanvas.style.removeProperty("width");
			renderingCanvas.style.removeProperty("height");
		}
		pixelScaling = 1;
	}
}

